
/* Chrysalide - Outil d'analyse de fichiers binaires
 * offset.c - constitution d'un décallage positif ou négatif
 *
 * Copyright (C) 2014-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "offset.h"


#include "../../../operand-int.h"
#include "../../../sharing/manager.h"
#include "../../../../common/sort.h"



/* --------------------- MANIPULATION D'OPERANDES INDIVIDUELLES --------------------- */


/* Définition d'un opérande visant à constituer un décallage relatif ARMv7 (instance) */
struct _GArmV7OffsetOperand
{
    GArchOperand parent;                    /* Instance parente            */

    bool positive;                          /* Sens du décallage           */
    GArchOperand *value;                    /* Valeur du décallage         */

};


/* Définition d'un opérande visant à constituer un décallage relatif ARMv7 (classe) */
struct _GArmV7OffsetOperandClass
{
    GArchOperandClass parent;               /* Classe parente              */

};


/* Initialise la classe des décallages relatifs ARMv7. */
static void g_armv7_offset_operand_class_init(GArmV7OffsetOperandClass *);

/* Initialise une instance de décallage relatif ARMv7. */
static void g_armv7_offset_operand_init(GArmV7OffsetOperand *);

/* Supprime toutes les références externes. */
static void g_armv7_offset_operand_dispose(GArmV7OffsetOperand *);

/* Procède à la libération totale de la mémoire. */
static void g_armv7_offset_operand_finalize(GArmV7OffsetOperand *);

/* Initialise un nouvel objet partagé avec des informations. */
static bool g_armv7_offset_operand_apply_template(GArmV7OffsetOperand *, const GArmV7OffsetOperand *);

/* Réalise une copie minimale d'un contenu partagé. */
static void g_armv7_offset_operand_define_template(const GArmV7OffsetOperand *, GArmV7OffsetOperand *);

/* Compare un opérande avec un autre. */
static int g_armv7_offset_operand_compare(const GArmV7OffsetOperand *, const GArmV7OffsetOperand *);

/* Traduit un opérande en version humainement lisible. */
static void g_armv7_offset_operand_print(const GArmV7OffsetOperand *, GBufferLine *, AsmSyntax);



/* -------------------------- PARTAGES DE CONTENUS UNIQUES -------------------------- */


/* Gestionnaire des partages d'instances */
static GShareManager *_armv7_offset_operand_manager = NULL;


/* Fournit le gestionnaire de partages attribué à un type. */
static GShareManager *get_armv7_offset_operand_share_manager(void);



/* ---------------------------------------------------------------------------------- */
/*                       MANIPULATION D'OPERANDES INDIVIDUELLES                       */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini par la GLib pour un décallage relatif ARMv7. */
G_DEFINE_TYPE(GArmV7OffsetOperand, g_armv7_offset_operand, G_TYPE_ARCH_OPERAND);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des décallages relatifs ARMv7.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_offset_operand_class_init(GArmV7OffsetOperandClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GArchOperandClass *operand;             /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);
    operand = G_ARCH_OPERAND_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_armv7_offset_operand_dispose;
    object->finalize = (GObjectFinalizeFunc)g_armv7_offset_operand_finalize;

    operand->get_manager = (get_operand_manager_fc)get_armv7_offset_operand_share_manager;

    operand->apply_template = (apply_operand_template_fc)g_armv7_offset_operand_apply_template;
    operand->define_template = (define_operand_template_fc)g_armv7_offset_operand_define_template;
    operand->free_template = (free_operand_template_fc)NULL;

    operand->compare = (operand_compare_fc)g_armv7_offset_operand_compare;
    operand->print = (operand_print_fc)g_armv7_offset_operand_print;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance à initialiser.                            *
*                                                                             *
*  Description : Initialise une instance de décallage relatif ARMv7.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_offset_operand_init(GArmV7OffsetOperand *operand)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_offset_operand_dispose(GArmV7OffsetOperand *operand)
{
    g_object_unref(G_OBJECT(operand->value));

    G_OBJECT_CLASS(g_armv7_offset_operand_parent_class)->dispose(G_OBJECT(operand));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_offset_operand_finalize(GArmV7OffsetOperand *operand)
{
    G_OBJECT_CLASS(g_armv7_offset_operand_parent_class)->finalize(G_OBJECT(operand));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand  = objet partagé à initialiser.                      *
*                template = information à utiliser pour la mise en place.     *
*                                                                             *
*  Description : Initialise un nouvel objet partagé avec des informations.    *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_armv7_offset_operand_apply_template(GArmV7OffsetOperand *operand, const GArmV7OffsetOperand *template)
{
    g_armv7_offset_operand_define_template(template, operand);

    g_object_ref(G_OBJECT(operand->value));

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand  = objet partagé à consulter.                        *
*                template = informations à retrouver intégralement.           *
*                                                                             *
*  Description : Réalise une copie minimale d'un contenu partagé.             *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_offset_operand_define_template(const GArmV7OffsetOperand *operand, GArmV7OffsetOperand *template)
{
    template->positive = operand->positive;

    template->value = operand->value;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : a = premier opérande à consulter.                            *
*                b = second opérande à consulter.                             *
*                                                                             *
*  Description : Compare un opérande avec un autre.                           *
*                                                                             *
*  Retour      : Bilan de la comparaison.                                     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int g_armv7_offset_operand_compare(const GArmV7OffsetOperand *a, const GArmV7OffsetOperand *b)
{
    int result;                             /* Bilan à faire remonter      */

    result = sort_boolean(a->positive, b->positive);
    if (result != 0) goto gaooc_done;

    result = g_arch_operand_compare(a->value, b->value);

 gaooc_done:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à traiter.                                *
*                line    = ligne tampon où imprimer l'opérande donné.         *
*                syntax  = type de représentation demandée.                   *
*                                                                             *
*  Description : Traduit un opérande en version humainement lisible.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_offset_operand_print(const GArmV7OffsetOperand *operand, GBufferLine *line, AsmSyntax syntax)
{
    if (!operand->positive)
        g_buffer_line_append_text(line, BLC_ASSEMBLY, "-", 1, RTT_KEY_WORD, NULL);

    g_arch_operand_print(operand->value, line, syntax);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : positive = indique si la quantité doit être ajoutée ou non.  *
*                value    = valeur du décallage à appliquer.                  *
*                                                                             *
*  Description : Crée un décallage selon un sens et une valeur donnés.        *
*                                                                             *
*  Retour      : Opérande mis en place.                                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchOperand *g_armv7_offset_operand_new(bool positive, GArchOperand *value)
{
    GSharedInstance *result;                /* Structure à retourner       */
    GArmV7OffsetOperand template;           /* Transport d'informations    */

    template.positive = positive;
    template.value = value;

    result = g_share_manager_build(_armv7_offset_operand_manager, (GSharedInstance *)&template);

    return G_ARCH_OPERAND(result);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à consulter.                              *
*                                                                             *
*  Description : Indique le sens du décallage représenté.                     *
*                                                                             *
*  Retour      : Indication d'ajout ou de retrait.                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_armv7_offset_operand_is_positive(const GArmV7OffsetOperand *operand)
{
    return operand->positive;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à consulter.                              *
*                                                                             *
*  Description : Founit la valeur utilisée pour un décallage.                 *
*                                                                             *
*  Retour      : Opérande en place.                                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchOperand *g_armv7_offset_operand_get_value(const GArmV7OffsetOperand *operand)
{
    GArchOperand *result;                   /* Instance à retourner        */

    result = operand->value;

    g_object_ref(G_OBJECT(result));

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                            PARTAGES DE CONTENUS UNIQUES                            */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit le gestionnaire de partages attribué à un type.      *
*                                                                             *
*  Retour      : Gestionnaire de partages en place.                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GShareManager *get_armv7_offset_operand_share_manager(void)
{
    return _armv7_offset_operand_manager;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Met en place les mécanismes de partage des décallages ARMv7. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool init_armv7_offset_operand_sharing(void)
{
    _armv7_offset_operand_manager = g_share_manager_new(G_TYPE_ARMV7_OFFSET_OPERAND);

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Imprime des statistiques quant aux partages dans l'archi.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/
#ifdef DEBUG_DUMP_STATS
void dump_armv7_offset_operand_share_stats(void)
{
    g_share_manager_dump_stats(_armv7_offset_operand_manager);

}
#endif


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Supprime les mécanismes de partage des décallages ARMv7.     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void exit_armv7_offset_operand_sharing(void)
{
    g_object_unref(G_OBJECT(_armv7_offset_operand_manager));

}
