
/* Chrysalide - Outil d'analyse de fichiers binaires
 * registers.c - aides auxiliaires relatives aux registres ARMv7
 *
 * Copyright (C) 2014-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "register.h"


#include <stdio.h>


#include "../register-int.h"



/* ------------------------- ENCADREMENT DE REGISTRES BRUTS ------------------------- */


/* Représentation d'un registre ARMv7 (instance) */
struct _GArmV7Register
{
    GArmRegister parent;                    /* Instance parente            */

};


/* Représentation d'un registre ARMv7 (classe) */
struct _GArmV7RegisterClass
{
    GArmRegisterClass parent;               /* Classe parente              */

};


#define MAX_REGNAME_LEN 8


/* Initialise la classe des registres ARMv7. */
static void g_armv7_register_class_init(GArmV7RegisterClass *);

/* Initialise une instance de registre ARMv7. */
static void g_armv7_register_init(GArmV7Register *);

/* Supprime toutes les références externes. */
static void g_armv7_register_dispose(GArmV7Register *);

/* Procède à la libération totale de la mémoire. */
static void g_armv7_register_finalize(GArmV7Register *);

/* Réalise une copie minimale d'un contenu partagé. */
static void g_armv7_register_define_template(const GArmV7Register *, GArmV7Register *);

/* Traduit un registre en version humainement lisible. */
static void g_armv7_register_print(const GArmV7Register *, GBufferLine *, AsmSyntax);



/* -------------------------- PARTAGES DE CONTENUS UNIQUES -------------------------- */


/* Gestionnaire des partages d'instances */
static GShareManager *_armv7_register_manager = NULL;


/* Fournit le gestionnaire de partages attribué à un type. */
static GShareManager *get_armv7_register_share_manager(void);



/* ---------------------------------------------------------------------------------- */
/*                           ENCADREMENT DE REGISTRES BRUTS                           */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour une représentation d'un registre ARMv7. */
G_DEFINE_TYPE(GArmV7Register, g_armv7_register, G_TYPE_ARM_REGISTER);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des registres ARMv7.                    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_register_class_init(GArmV7RegisterClass *klass)
{
    GObjectClass *object_class;             /* Autre version de la classe  */
    GArchRegisterClass *reg_class;          /* Classe de haut niveau       */

    object_class = G_OBJECT_CLASS(klass);
    reg_class = G_ARCH_REGISTER_CLASS(klass);

    object_class->dispose = (GObjectFinalizeFunc/* ! */)g_armv7_register_dispose;
    object_class->finalize = (GObjectFinalizeFunc)g_armv7_register_finalize;

    reg_class->get_manager = (get_register_manager_fc)get_armv7_register_share_manager;

    reg_class->define_template = (define_register_template_fc)g_armv7_register_define_template;

    reg_class->print = (reg_print_fc)g_armv7_register_print;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : reg = instance à initialiser.                                *
*                                                                             *
*  Description : Initialise une instance de registre ARMv7.                   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_register_init(GArmV7Register *reg)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : reg = instance d'objet GLib à traiter.                       *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_register_dispose(GArmV7Register *reg)
{
    G_OBJECT_CLASS(g_armv7_register_parent_class)->dispose(G_OBJECT(reg));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : reg = instance d'objet GLib à traiter.                       *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_register_finalize(GArmV7Register *reg)
{
    G_OBJECT_CLASS(g_armv7_register_parent_class)->finalize(G_OBJECT(reg));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : reg      = objet partagé à consulter.                        *
*                template = informations à retrouver intégralement.           *
*                                                                             *
*  Description : Réalise une copie minimale d'un contenu partagé.             *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_register_define_template(const GArmV7Register *reg, GArmV7Register *template)
{
    GArmRegister *areg;                     /* Version parente #1          */
    GArmRegister *treg;                     /* Version parente #2          */

    areg = (GArmRegister *)reg;
    treg = (GArmRegister *)template;

    treg->index = areg->index;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : reg    = registre à transcrire.                              *
*                line   = ligne tampon où imprimer l'opérande donné.          *
*                syntax = type de représentation demandée.                    *
*                                                                             *
*  Description : Traduit un registre en version humainement lisible.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_register_print(const GArmV7Register *reg, GBufferLine *line, AsmSyntax syntax)
{
    char key[MAX_REGNAME_LEN];              /* Mot clef principal          */
    size_t klen;                            /* Taille de ce mot clef       */

    switch (G_ARM_REGISTER(reg)->index)
    {
        case 0 ... 12:
            klen = snprintf(key, MAX_REGNAME_LEN, "r%hhu", G_ARM_REGISTER(reg)->index);
            break;
        case 13:
            klen = snprintf(key, MAX_REGNAME_LEN, "sp");
            break;
        case 14:
            klen = snprintf(key, MAX_REGNAME_LEN, "lr");
            break;
        case 15:
            klen = snprintf(key, MAX_REGNAME_LEN, "pc");
            break;
        case 16:
            klen = snprintf(key, MAX_REGNAME_LEN, "cpsr");
            break;
        case 17:
            klen = snprintf(key, MAX_REGNAME_LEN, "spsr");
            break;
        default:
            klen = snprintf(key, MAX_REGNAME_LEN, "r??");
            break;
    }

    g_buffer_line_append_text(line, BLC_ASSEMBLY, key, klen, RTT_REGISTER, NULL);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : index = indice du registre correspondant.                    *
*                                                                             *
*  Description : Crée une réprésentation de registre ARMv7.                   *
*                                                                             *
*  Retour      : Adresse de la structure mise en place.                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArmV7Register *g_armv7_register_new(uint8_t index)
{
    GArmV7Register *result;                 /* Structure à retourner       */
    GArmRegister template;                  /* Transport d'informations    */

    template.index = index;

    result = G_ARMV7_REGISTER(g_share_manager_build(_armv7_register_manager, (GSharedInstance *)&template));

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                            PARTAGES DE CONTENUS UNIQUES                            */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit le gestionnaire de partages attribué à un type.      *
*                                                                             *
*  Retour      : Gestionnaire de partages en place.                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GShareManager *get_armv7_register_share_manager(void)
{
    return _armv7_register_manager;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Met en place les mécanismes de partage des registres ARMv7.  *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool init_armv7_register_sharing(void)
{
    _armv7_register_manager = g_share_manager_new(G_TYPE_ARMV7_REGISTER);

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Imprime des statistiques quant aux partages dans l'archi.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/
#ifdef DEBUG_DUMP_STATS
void dump_armv7_register_share_stats(void)
{
    g_share_manager_dump_stats(_armv7_register_manager);

}
#endif


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Supprime les mécanismes de partage des registres ARMv7.      *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void exit_armv7_register_sharing(void)
{
    g_object_unref(G_OBJECT(_armv7_register_manager));

}
